@@ -1,5 +1,6 @@
 /*
  Copyright (C) 2006-2007 M.A.L. Marques
+ Copyright (C) 2019 X. Andrade
 
  This Source Code Form is subject to the terms of the Mozilla Public
  License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -16,8 +17,9 @@
 #include <strings.h>
 #endif
 
-extern xc_func_info_type 
-  *xc_lda_known_funct[], 
+extern xc_func_info_type
+  *xc_lda_known_funct[],
+  *xc_hyb_lda_known_funct[],
   *xc_gga_known_funct[],
   *xc_hyb_gga_known_funct[],
   *xc_mgga_known_funct[],
@@ -46,7 +48,7 @@ int xc_functional_get_number(const char *name)
       break;
     }
   }
-  
+
   return key;
 }
 
@@ -63,7 +65,7 @@ char *xc_functional_get_name(int number)
     if(xc_functional_keys[ii].number == number) {
       /* return duplicated: caller has the responsibility to dealloc string.
          Do this the old way since strdup and strndup aren't C standard. */
-      p=malloc(strlen(xc_functional_keys[ii].name)+1);
+      p = (char *) libxc_malloc(strlen(xc_functional_keys[ii].name) + 1);
       strcpy(p,xc_functional_keys[ii].name);
       return p;
     }
@@ -85,6 +87,14 @@ int xc_family_from_id(int id, int *family, int *number)
     }
   }
 
+  for(ii=0; xc_hyb_lda_known_funct[ii]!=NULL; ii++){
+    if(xc_hyb_lda_known_funct[ii]->number == id){
+      if(family != NULL) *family = XC_FAMILY_HYB_LDA;
+      if(number != NULL) *number = ii;
+      return XC_FAMILY_HYB_LDA;
+    }
+  }
+
   /* or is it a GGA? */
   for(ii=0; xc_gga_known_funct[ii]!=NULL; ii++){
     if(xc_gga_known_funct[ii]->number == id){
@@ -94,7 +104,6 @@ int xc_family_from_id(int id, int *family, int *number)
     }
   }
 
-  /* or is it a hybrid GGA? */
   for(ii=0; xc_hyb_gga_known_funct[ii]!=NULL; ii++){
     if(xc_hyb_gga_known_funct[ii]->number == id){
       if(family != NULL) *family = XC_FAMILY_HYB_GGA;
@@ -112,7 +121,6 @@ int xc_family_from_id(int id, int *family, int *number)
     }
   }
 
-  /* or is it a hybrid meta GGA? */
   for(ii=0; xc_hyb_mgga_known_funct[ii]!=NULL; ii++){
     if(xc_hyb_mgga_known_funct[ii]->number == id){
       if(family != NULL) *family = XC_FAMILY_HYB_MGGA;
@@ -154,6 +162,10 @@ int xc_maximum_name_length()
 }
 
 /*------------------------------------------------------*/
+static int compare_int(const void *a, const void *b) {
+  return *(int *)a - *(int *) b;
+}
+
 void xc_available_functional_numbers(int *list)
 {
   int ii, N;
@@ -161,16 +173,74 @@ void xc_available_functional_numbers(int *list)
   for(ii=0;ii<N;ii++){
     list[ii]=xc_functional_keys[ii].number;
   }
+  /* Sort list by id */
+  qsort(list, N, sizeof(int), compare_int);
+}
+
+static int compare_func_names(const void *a, const void *b) {
+  int ia, ib;
+  int fama, famb;
+  int hyba, hybb;
+
+  ia = *(int *)a;
+  ib = *(int *)b;
+
+  /* First we sort by the family: LDAs, GGAs, meta-GGAs */
+  fama = xc_family_from_id(xc_functional_keys[ia].number, NULL, NULL);
+  famb = xc_family_from_id(xc_functional_keys[ib].number, NULL, NULL);
+  if(fama < famb)
+    return -1;
+  else if(fama > famb)
+    return 1;
+
+  /* Then we sort by hybrid type: non-hybrids go first */
+  hyba = (strncmp(xc_functional_keys[ia].name, "hyb_", 4) == 0);
+  hybb = (strncmp(xc_functional_keys[ib].name, "hyb_", 4) == 0);
+  if(!hyba && hybb)
+    return -1;
+  else if(hyba && !hybb)
+    return 1;
+
+  /* Last we sort by name */
+  return strcmp(xc_functional_keys[ia].name, xc_functional_keys[ib].name);
+}
+
+void xc_available_functional_numbers_by_name(int *list)
+{
+  int ii, N;
+
+  /* Arrange list of functional IDs by name */
+  N=xc_number_of_functionals();
+  for(ii=0;ii<N;ii++) {
+    list[ii]=ii;
+  }
+  qsort(list, N, sizeof(int), compare_func_names);
+  /* Map the internal list to functional IDs */
+  for(ii=0;ii<N;ii++){
+    list[ii]=xc_functional_keys[list[ii]].number;
+  }
 }
 
 void xc_available_functional_names(char **list)
 {
   int ii, N;
+  int *idlist;
 
+  /* Arrange list of functional IDs by name */
   N=xc_number_of_functionals();
+  idlist=(int *) libxc_malloc(N*sizeof(int));
   for(ii=0;ii<N;ii++) {
-    strcpy(list[ii],xc_functional_keys[ii].name);
+    idlist[ii]=ii;
   }
+  qsort(idlist, N, sizeof(int), compare_func_names);
+
+  /* Now store the correctly ordered output */
+  for(ii=0;ii<N;ii++) {
+    strcpy(list[ii],xc_functional_keys[idlist[ii]].name);
+  }
+
+  /* Deallocate work array */
+  libxc_free(idlist);
 }
 
 /*------------------------------------------------------*/
@@ -178,10 +248,36 @@ xc_func_type *xc_func_alloc()
 {
   xc_func_type *func;
 
-  func = (xc_func_type *) malloc (sizeof (xc_func_type));
+  func = (xc_func_type *) libxc_malloc (sizeof (xc_func_type));
   return func;
 }
 
+/*------------------------------------------------------*/
+void xc_func_nullify(xc_func_type *func)
+{
+  assert(func != NULL);
+
+  func->info       = NULL;
+  func->nspin      = XC_UNPOLARIZED;
+
+  func->n_func_aux = 0;
+  func->func_aux   = NULL;
+  func->mix_coef   = NULL;
+
+  func->cam_omega  = 0.0;
+  func->cam_alpha  = 0.0;
+  func->cam_beta   = 0.0;
+
+  func->nlc_b = func->nlc_C = 0.0;
+
+  func->params     = NULL;
+
+  func->dens_threshold  = 0.0;
+  func->zeta_threshold  = 0.0;
+  func->sigma_threshold = 0.0;
+  func->tau_threshold   = 0.0;
+}
+
 /*------------------------------------------------------*/
 int xc_func_init(xc_func_type *func, int functional, int nspin)
 {
@@ -190,93 +286,71 @@ int xc_func_init(xc_func_type *func, int functional, int nspin)
   assert(func != NULL);
   assert(nspin==XC_UNPOLARIZED || nspin==XC_POLARIZED);
 
+  xc_func_nullify(func);
+
   /* initialize structure */
   func->nspin       = nspin;
 
-  func->params     = NULL;
-
-  func->n_func_aux = 0;
-  func->func_aux   = NULL;
-  func->mix_coef   = NULL;
-  func->cam_omega = func->cam_alpha = func->cam_beta = 0.0;
-  func->nlc_b = func->nlc_C = 0.0;
+  // we have to make a copy because the *_known_funct arrays live in
+  // host memory (libxc_malloc instead returns memory than can be read
+  // from GPU and CPU).
+  xc_func_info_type * finfo = (xc_func_info_type *) libxc_malloc(sizeof(xc_func_info_type));
 
   switch(xc_family_from_id(functional, NULL, &number)){
   case(XC_FAMILY_LDA):
-    func->info = xc_lda_known_funct[number];
+    *finfo = *xc_lda_known_funct[number];
+    internal_counters_set_lda(func->nspin, &(func->dim));
     break;
 
+  case(XC_FAMILY_HYB_LDA):
+    *finfo = *xc_hyb_lda_known_funct[number];
+    internal_counters_set_lda(func->nspin, &(func->dim));
+    break;
+    
   case(XC_FAMILY_GGA):
-    func->info = xc_gga_known_funct[number];
+    *finfo = *xc_gga_known_funct[number];
+    internal_counters_set_gga(func->nspin, &(func->dim));
     break;
 
   case(XC_FAMILY_HYB_GGA):
-    func->info = xc_hyb_gga_known_funct[number];
+    *finfo = *xc_hyb_gga_known_funct[number];
+    internal_counters_set_gga(func->nspin, &(func->dim));
     break;
 
   case(XC_FAMILY_MGGA):
-    func->info = xc_mgga_known_funct[number];
+    *finfo = *xc_mgga_known_funct[number];
+    internal_counters_set_mgga(func->nspin, &(func->dim));
     break;
 
   case(XC_FAMILY_HYB_MGGA):
-    func->info = xc_hyb_mgga_known_funct[number];
+    *finfo = *xc_hyb_mgga_known_funct[number];
+    internal_counters_set_mgga(func->nspin, &(func->dim));
     break;
 
   default:
     return -2; /* family not found */
   }
 
-  /* setup internal counters */
-  switch(xc_family_from_id(functional, NULL, &number)){
-  case(XC_FAMILY_MGGA):
-  case(XC_FAMILY_HYB_MGGA):
-    func->n_tau  = func->n_vtau = func->nspin;
-    func->n_lapl = func->n_vlapl = func->nspin;
-    if(func->nspin == XC_UNPOLARIZED){
-      func->n_v2tau2 = func->n_v2lapl2 = 1;
-      func->n_v2rhotau = func->n_v2rholapl = func->n_v2lapltau = 1;
-      func->n_v2sigmatau = func->n_v2sigmalapl = 1;
-    }else{
-      func->n_v2tau2 = func->n_v2lapl2 = 3;
-      func->n_v2rhotau = func->n_v2rholapl = func->n_v2lapltau = 4;
-      func->n_v2sigmatau = func->n_v2sigmalapl = 6;
-    }
+  func->info = finfo;
 
-  case(XC_FAMILY_GGA):
-  case(XC_FAMILY_HYB_GGA):
-    if(func->nspin == XC_UNPOLARIZED){
-      func->n_sigma  = func->n_vsigma = 1;
-      func->n_v2rhosigma  = func->n_v2sigma2 = 1;
-      func->n_v3rho2sigma = func->n_v3rhosigma2 = func->n_v3sigma3 = 1;
-    }else{
-      func->n_sigma      = func->n_vsigma = 3;
-      func->n_v2rhosigma = func->n_v2sigma2 = 6;
+  /* this is initialized for each functional from the info */
+  func->dens_threshold = func->info->dens_threshold;
+  /* the density and sigma cutoffs should be connected, especially in
+     the case of kinetic energy functionals. This is the correct
+     scaling */
+  func->sigma_threshold = pow(func->info->dens_threshold, 4.0/3.0);
 
-      func->n_v3rho2sigma = 9;
-      func->n_v3rhosigma2 = 12;
-      func->n_v3sigma3    = 10;
-    }
-
-  case(XC_FAMILY_LDA):
-    func->n_rho = func->n_vrho = func->nspin;
-    func->n_zk  = 1;
-    if(func->nspin == XC_UNPOLARIZED){
-      func->n_v2rho2 = func->n_v3rho3 = 1;
-    }else{
-      func->n_v2rho2 = 3;
-      func->n_v3rho3 = 4;
-    }
-  }
+  /* these are reasonable defaults */
+  func->zeta_threshold  = DBL_EPSILON;
+  func->tau_threshold   = 1e-20;
 
   /* see if we need to initialize the functional */
   if(func->info->init != NULL)
     func->info->init(func);
 
   /* see if we need to initialize the external parameters */
-  if(func->info->n_ext_params > 0)
-    func->info->set_ext_params(func, NULL);
-
-  func->dens_threshold = func->info->dens_threshold;
+  if(func->info->ext_params.n > 0)
+    func->info->ext_params.set(func, NULL);
 
   return 0;
 }
@@ -297,30 +371,28 @@ void xc_func_end(xc_func_type *func)
 
     for(ii=0; ii<func->n_func_aux; ii++){
       xc_func_end(func->func_aux[ii]);
-      free(func->func_aux[ii]);
+      libxc_free(func->func_aux[ii]);
     }
-    free(func->func_aux);
-    func->n_func_aux = 0;
+    libxc_free(func->func_aux);
   }
 
-  if(func->mix_coef != NULL){
-    free(func->mix_coef);
-    func->mix_coef = NULL;
-  }
+  /* deallocate coefficients for mixed functionals */
+  if(func->mix_coef != NULL)
+    libxc_free(func->mix_coef);
 
   /* deallocate any used parameter */
-  if(func->params != NULL){
-    free(func->params);
-    func->params = NULL;
-  }
+  if(func->params != NULL)
+    libxc_free(func->params);
 
-  func->info = NULL;  
+  libxc_free((void *) func->info);
+
+  xc_func_nullify(func);
 }
 
 /*------------------------------------------------------*/
 void  xc_func_free(xc_func_type *p)
 {
-  free(p);
+  libxc_free(p);
 }
 
 /*------------------------------------------------------*/
@@ -330,41 +402,87 @@ const xc_func_info_type *xc_func_get_info(const xc_func_type *p)
 }
 
 /*------------------------------------------------------*/
-void xc_func_set_dens_threshold(xc_func_type *p, double dens_threshold)
+void xc_func_set_dens_threshold(xc_func_type *p, double t_dens)
 {
   int ii;
 
-  p->dens_threshold = dens_threshold;
+  if(t_dens  > 0.0)
+    p->dens_threshold = t_dens;
 
   for(ii=0; ii<p->n_func_aux; ii++) {
-    xc_func_set_dens_threshold(p->func_aux[ii], dens_threshold);
+    xc_func_set_dens_threshold(p->func_aux[ii], t_dens);
+  }
+}
+/*------------------------------------------------------*/
+void xc_func_set_zeta_threshold(xc_func_type *p, double t_zeta)
+{
+  int ii;
+
+  if(t_zeta  > 0.0)
+    p->zeta_threshold = t_zeta;
+
+  for(ii=0; ii<p->n_func_aux; ii++) {
+    xc_func_set_zeta_threshold(p->func_aux[ii], t_zeta);
+  }
+}
+/*------------------------------------------------------*/
+void xc_func_set_sigma_threshold(xc_func_type *p, double t_sigma)
+{
+  int ii;
+
+  if(t_sigma  > 0.0)
+    p->sigma_threshold = t_sigma;
+
+  for(ii=0; ii<p->n_func_aux; ii++) {
+    xc_func_set_sigma_threshold(p->func_aux[ii], t_sigma);
+  }
+}
+/*------------------------------------------------------*/
+void xc_func_set_tau_threshold(xc_func_type *p, double t_tau)
+{
+  int ii;
+
+  if(t_tau  > 0.0)
+    p->tau_threshold = t_tau;
+
+  for(ii=0; ii<p->n_func_aux; ii++) {
+    xc_func_set_tau_threshold(p->func_aux[ii], t_tau);
   }
 }
 
 /*------------------------------------------------------*/
-void xc_func_set_ext_params(xc_func_type *p, double *ext_params)
+/* get/set external parameters                          */
+void
+xc_func_set_ext_params(xc_func_type *p, const double *ext_params)
 {
-  assert(p->info->n_ext_params > 0);
-  p->info->set_ext_params(p, ext_params);
+  assert(p->info->ext_params.n > 0);
+  p->info->ext_params.set(p, ext_params);
 }
 
-/* returns the mixing coefficient for the hybrid GGAs */
-double xc_hyb_exx_coef(const xc_func_type *p)
+void
+xc_func_set_ext_params_name(xc_func_type *p, const char *name, double par)
 {
-  assert(p!=NULL);
- 
-  return p->cam_alpha;
+  int ii;
+  double *ext_params;
+  int name_found=0;
+
+  assert(p != NULL && p->info->ext_params.n > 0);
+
+  ext_params = (double *) libxc_malloc(p->info->ext_params.n*sizeof(double));
+  for(ii=0; ii<p->info->ext_params.n; ii++){
+    if(strcmp(p->info->ext_params.names[ii], name) == 0) {
+      ext_params[ii] = par;
+      name_found=1;
+    } else {
+      ext_params[ii] = XC_EXT_PARAMS_DEFAULT;
+    }
+  }
+  xc_func_set_ext_params(p, ext_params);
+  libxc_free(ext_params);
+  /* Check that we found the parameter */
+  assert(name_found);
 }
 
-/* returns the CAM parameters for screened hybrids */
-void xc_hyb_cam_coef(const xc_func_type *p, double *omega, double *alpha, double *beta)
-{
-  assert(p!=NULL);
-
-  *omega = p->cam_omega;
-  *alpha = p->cam_alpha;
-  *beta  = p->cam_beta;
-}
 
 /* returns the NLC parameters */
 void xc_nlc_coef(const xc_func_type *p, double *nlc_b, double *nlc_C)
